#include <sys/types.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"

/**
 * filter callback function (from PKGDIR)
 * since directory name in PKGDIR is the same as names of all packages
 * @return 0	if not necessary
 * 	   1	if necessary
 */
static int need_handle(struct dirent *dp) {
  if (dp->d_type != DT_DIR) {
    return 0;
  } else if (dp->d_namlen <= 0) {
    return 0;
  } else if (dp->d_name[0] == '.') {
    /**
     * Don't handle . and ..
     */
    return 0;
  }
  return 1;
}

void process_entry(DBI *dbp, struct dirent *dp) {
  int ret;
  DBI_UNIT key, data;

  ret = dbi_get(dbp, &key, &data, dp->d_name, dp->d_namlen);
  if (ret == DB_NOTFOUND) {
    char *content_file;
    FILE *cfp;
    char fbuf[1024];
    const char compare_start[] = "@comment ORIGIN:";
    const int cs_len = sizeof(compare_start)/sizeof(compare_start[0]) - 1;

    asprintf(&content_file, PKGDIR "%.*s/+CONTENTS", dp->d_namlen, dp->d_name);
    cfp = fopen(content_file, "r");
    while (fgets(fbuf, sizeof(fbuf), cfp)) {
      if (strncmp(fbuf, compare_start, cs_len) == 0) {
	char *p_fbuf = strchr(&fbuf[cs_len], '\n');
	int fbuf_cslen;
	if (p_fbuf) {
	  *p_fbuf = '\0';
	}
	fbuf_cslen = strlen(&fbuf[cs_len]);

	ret = dbi_get(dbp, &key, &data, &fbuf[cs_len], fbuf_cslen);
	if (ret == DB_NOTFOUND) {
	  char moved_loc[512] = "\0";
	  if (!get_moved_port(&fbuf[cs_len], fbuf_cslen, moved_loc, sizeof(moved_loc))) {
	    /**
	     * get_moved_port
	     * 0 Found
	     */
	    printf("%-32.*s\t", dp->d_namlen, dp->d_name);
	    if (0) {
	      printf("Not found in db");
	    } else if (moved_loc[0]) {
	      printf("Port move to %s", moved_loc);
	    } else {
	      printf("Port removed from %.*s", fbuf_cslen, &fbuf[cs_len]);
	    }
	  } else {
	    /**
	     * get_moved_port
	     * -1 Not found
	     */
	    printf("\033[33m%-32.*s\tis not in " PORTSDIR "/INDEX\033[m", dp->d_namlen, dp->d_name);
	  }
	  printf("\n");

	} else {
	  char *data_p;
	  char *data_sp;
	  char *show_p;
	  char *p1, *p2;
	  /**
	   * show_p: dp->d_name 's c-string version
	   * data_p: data.data 's c-string version
	   */
	  asprintf(&show_p, "%.*s", dp->d_namlen, dp->d_name);
	  asprintf(&data_p, "%.*s", data.DU_LEN, (const char *)data.DU_DATA);

	  /**
	   * If string data_p contains |, data_p is terminated at |
	   * (If no | in data_p, still hold since it's a c-string version
	   * where data_p is port-name
	   *       show_p is pkg-name
	   */
	  data_sp = strchr(data_p, '|');
	  if (data_sp) {
	    *data_sp = '\0';
	  }
	  p1 = strrchr(show_p, '-');
	  p2 = strrchr(data_p, '-');
	  if (p1 && p2) {
	    /**
	     * Only display show_p's name without version
	     */
	    *p1 = '\0';
	    /**
	     * Not important
	     */
	    *p2 = '\0';
	    p1++;
	    p2++;
	    if (strcmp(p1, p2) != 0) {
	      printf("%-32s\t%-16s\t%s\n", show_p, p1, p2);
	    }
	  } else {
	    if (strcmp(show_p, data_p) != 0) {
	      printf("%-32s\t%s\n", show_p, data_p);
	    }
	  }
	  free(data_p);
	  free(show_p);
	}
	break;
      }
    }
    fclose(cfp);
    free(content_file);
  }
}

int main(int argc, char *argv[]) {
	DBI *dbp;
	if (dbi_init(&dbp) < 0) {
	  fprintf(stderr, "BDB Init error\n");
	  return -1;
	}

#if SORT_PACKAGE
	struct dirent **namelist;
	int n_package = scandir(PKGDIR, &namelist, need_handle, alphasort);
	int i;

	for (i = 0; i < n_package; i++) {
	  process_entry(dbp, namelist[i]);
	  free(namelist[i]);
	}
	free(namelist);
#else
	DIR *dirp;
	struct dirent *dp;

	dirp = opendir(PKGDIR);
	while ((dp = readdir(dirp)) != NULL) {
	  /**
	   * We only handle directory name(pkg-name)
	   * The remainding part will only handle the directories in /var/db/pkg
	   */
	  if (!need_handle(dp)) {
	    continue;
	  }
	  process_entry(dbp, dp);
	}

	if (dirp != NULL) {
	  closedir(dirp);
	}
#endif

	dbi_close(dbp);

	return 0;
}
